'
' DTMF decoder using a 600Hz to 2KHz FIR bandpass Filter and an FFT.
' For use with Proton24 version 1.0.2.3 onwards
' Written by Les Johnson 11-05-14
'
' For accurate operation, the signal on AN0 should be no more than 1.0 Volt and have a DC offset of 1.7 Volts.
'
' DTMF keypad and frequencies(in Hz)
'
'  --------------------------------
' |    | 1209 | 1336 | 1477 | 1633 |
' |----|---------------------------|
' |697 |   1  |   2  |   3  |   A  |
' |----|---------------------------|
' |770 |   4  |   5  |   6  |   B  |
' |----|---------------------------|
' |852 |   7  |   8  |   9  |   C  |
' |----|---------------------------|
' |941 |   *  |   0  |   #  |   D  |
'  --------------------------------
'
    Device = 33FJ128GP802
    Declare Xtal = 79.23

    Declare Hserial_Baud = 115200               ' UART1 baud rate
    Declare Hrsout1_Pin = PORTB.3               ' Select which pin is to be used for TX with USART1

    Include "FIR.inc"                           ' Load the FIR procedures into the program
    Include "FFT.inc"                           ' Load the FFT procedures into the program

$ifndef False
    $define False 0
$endif
$ifndef True
    $define True 1
$endif
'
' Timer1 interrupt (ADC sample) rate calculator
'
    $define cMicroSeconds $eval (1000000 / 8000)                            ' Interrupt rate (in us). Sample rate is (1000000 / Frequency (in Hz))
    $define cTweak_Value 36                                                 ' Added tweak value for rounding errors
    $define cPR1_Value $eval (cMicroSeconds * (_xtal / 2)) + cTweak_Value   ' Calculate the actual value to place into the PR1 SFR

    $define cFFT_BufferSize 256                                             ' 256 pairs of complex (Re, Im) values, resulting in 128 FFT bins
    $define cFFT_BinSize $eval (cFFT_BufferSize / 2)                        ' The amount fo resulting FFT bins

    $define cFIR_BufferSize 128                                             ' The size of the FIR filter's YRAM buffer
    $define cFIR_Taps 128                                                   ' The amount of filter taps (the coefficient table)
'
' Coefficient table addressable in PSV space
' The values below are for a 600Hz to 2KHz band pass filter using a Hamming window
'
    Dim cCOEFF As PSV = As Word $0019, $0000, $0016, $002A, $0000, $FFD0,_
                                $FFE3, $0000, $FFD7, $FFB0, $FFF0, $0042,_
                                $0025, $FFF8, $004D, $00A0, $0037, $FFA9,_
                                $FFDD, $0024, $FF80, $FEDB, $FF7E, $005F,_
                                $0000, $FF90, $00B5, $01E0, $00F8, $FFAE,_
                                $005E, $0114, $FF31, $FD34, $FE69, $0030,_
                                $FEEC, $FDBA, $00A7, $03DE, $0250, $FFF8,_
                                $0245, $0462, $0000, $FAED, $FCF4, $000D,_
                                $FBCF, $F7D5, $FE64, $0698, $03AF, $FF3F,_
                                $07B5, $105E, $05C7, $F672, $FBE3, $04B5,_
                                $ECCC, $CA47, $DF1C, $2C78, $599A, $2C78,_
                                $DF1C, $CA47, $ECCC, $04B5, $FBE3, $F672,_
                                $05C7, $105E, $07B5, $FF3F, $03AF, $0698,_
                                $FE64, $F7D5, $FBCF, $000D, $FCF4, $FAED,_
                                $0000, $0462, $0245, $FFF8, $0250, $03DE,_
                                $00A7, $FDBA, $FEEC, $0030, $FE69, $FD34,_
                                $FF31, $0114, $005E, $FFAE, $00F8, $01E0,_
                                $00B5, $FF90, $0000, $005F, $FF7E, $FEDB,_
                                $FF80, $0024, $FFDD, $FFA9, $0037, $00A0,_
                                $004D, $FFF8, $0025, $0042, $FFF0, $FFB0,_
                                $FFD7, $0000, $FFE3, $FFD0, $0000, $002A,_
                                $0016, $0000, $0019
'
' Variables used for the FIR filter
'
    Dim FIR_wIndex As Word                                  ' FIR filter's input buffer index
    Dim FIR_wSamples[cFIR_BufferSize] As Word YRAM          ' FIR filter's input buffer, must be in Y RAM
'
' Variables used by FFT
'
    Dim FFT_wIndex As Word                                  ' FFT's buffer index
    Dim FFT_wSamples[cFFT_BufferSize * 2] As Word YRAM      ' Y RAM space is required by FFT routine
    Dim FFT_tBufferReady As Bit                             ' True if the FFT buffer is full (no more allowed until cleared)
'
' Variables used for the DTMF decode
'
    Dim DTMF_bKeyValue As Byte                              ' Holds the key value from the DTMF decoder
    Dim DTMF_wZeroCounter As Word                           ' Inter-character counter

'----------------------------------------------------------------------------------------------------
' Setup the ADC and initialise an ADC interrupt
' Input     : None
' Output    : None
' Notes     : The procedure also encompasses the interrupt handler
'
Proc ADC_Start()
    Input PORTA.0                           ' Make AN0 an input
    AD1CON2 = 0                             ' AVdd, AVss, MUXA only
    AD1PCFGLbits_PCFG0 = 0                  ' Analogue input on AN0
    AD1CHS0 = 0                             ' Configure AN0 as the ADC input
    AD1CON2 = $0000
    AD1CON3 = $021A                         ' Sampling time = 3 * Tad, minimum Tad selected
    AD1CON1 = %1000001111100010
'              ||||||||||||||||______________ Bit-0 : No ADC conversion yet
'              |||||||||||||||_______________ Bit-1 : ADC Sample and Hold amplifiers are holding
'              ||||||||||||||________________ Bit-2 : Sampling begins immediately after last conversion. SAMP bit is auto-set.
'              |||||||||||||_________________ Bit-3 : Samples multiple channels individually in sequence
'              ||||||||||||__________________ Bit-4 : Sample Clock Source Group Bit
'              |||||||||||___________________ Bit-5 :\
'              ||||||||||____________________ Bit-6 :| Internal Counter ends sampling and starts conversion (auto-convert)
'              |||||||||_____________________ Bit-7 :/
'              ||||||||______________________ Bit-8 :\ ADC is using signed fractional (DOUT = sddddddddddd)
'              |||||||_______________________ Bit-9 :/
'              ||||||________________________ Bit-10: 12-Bit, 1-channel ADC operation
'              |||||_________________________ Bit-11: Unimplemented
'              ||||__________________________ Bit-12: DMA buffers are written in Scatter/Gather mode
'              |||___________________________ Bit-13: Continue module operation in Idle mode
'              ||____________________________ Bit-14: Unimplemented
'              |_____________________________ Bit-15: Enable the ADC

    IEC0bits_AD1IE = 1                      ' ADC interrupt enabled
    AD1CON1bits_ADON = 1                    ' Enable the ADC
    AD1CON1bits_SAMP = 1                    ' Start a sample
'
'-------------------------------------------
' ADC interrupt handler
' Input     : None
' Output    : FFT_wSamples holds the FFT bins
' Notes     : Filters the ADC result and places it in the real and imaginary parts of the FFT buffer
'
Isr ADC1Interrupt
    If FFT_tBufferReady = False Then                        ' Is the FFT samples filter full?
        FIR_wSamples[FIR_wIndex] = ADCBUF0                  ' No. So get an ADC sample into the filter's buffer
        FFT_wSamples[FFT_wIndex] = FIR_Filter(FIRStruct, FIR_wIndex) ' Place the filtered value into the FFT array
        '
        ' Manage the FIR filter's buffer
        '
        Inc FIR_wIndex                                      ' Increment the samples buffer index
        If FIR_wIndex >= cFIR_BufferSize Then               ' Have we reached the end of the buffer?
            FIR_wIndex = 0                                  ' Yes. So reset it
        EndIf
        '
        ' Manage the FFT buffer
        '
        Inc FFT_wIndex                                      ' Increment the samples buffer index
        If FFT_wIndex >= cFFT_BufferSize Then               ' Have we reached the end of the FFT buffer?
            FFT_wIndex = 0                                  ' Yes. So reset it
            FFT_tBufferReady = True                         ' Signal to the main program that we have values in the buffer
        EndIf
    EndIf
    IFS0bits_AD1IF = 0                                      ' Clear the ADC interrupt flag
EndIsr
EndProc

'----------------------------------------------------------------------------------------------------
' Initialise a Timer1 interrupt
' Input     : None
' Output    : None
' Notes     : The procedure also encompasses the Interrupt handler
'
Proc Timer1_Start()
    T1CON = 0                                   ' Reset the T1CON SFR
    T1CONbits_TCKPS0 = 0                        ' \
    T1CONbits_TCKPS1 = 0                        ' / Timer1 Prescaler 1:1
    PR1 = cPR1_Value                            ' ADC Sampling rate
                                                ' Set the match cPR1_Value (will cause an interrupt when this value is reached)
    TMR1 = 0                                    ' Reset Timer1 value
    IFS0bits_T1IF = 0                           ' Clear the Timer1 interrupt flag
    IPC0bits_T1IP0 = 1                          ' \
    IPC0bits_T1IP1 = 1                          ' | High Priority Timer1 interrupt
    IPC0bits_T1IP2 = 1                          ' /
    T1CONbits_TON = 1                           ' Turn on Timer1
    IEC0bits_T1IE = 1                           ' Enable the Timer1 interrupt
'
'-----------------------------------------------
' Timer1 Interrupt handler
' Input     : None
' Output    : None
' Notes     : Triggers an ADC interrupt
'
Isr T1Interrupt
    If AD1CON1bits_DONE = 1 Then                ' Is the ADC busy?
        AD1CON1bits_SAMP = 1                    ' No. So start a new ADC sample
    EndIf
    IFS0bits_T1IF = 0                           ' Clear the Timer1 interrupt flag
EndIsr
EndProc

'-------------------------------------------------------------------------
' Find the highest row peak within the FFT array
' Input     : FFT_wSamples
' Output    : Returns the bin that contains the largest value
' Notes     : Only scans the first half of the array because this is where the row frequencies are
'
Proc DTMF_FindRowBin(), Byte
    Dim wTemp As Word
    Dim wIndex As Word
    Dim wLargest As Word = 2

    wIndex = 10                             ' Start at bin 10 (in case there is noise)
    Result = 0                              ' Default return if no bin found
    Repeat                                  ' Create loop to scan the bins
        wTemp = FFT_wSamples[wIndex]        ' Get a bin's value
        If wTemp > wLargest Then            ' Is this value greater than the largest found so far?
            wLargest = wTemp                ' Yes. So make this value the largest so far
            Result = wIndex                 ' Make this value the result
        EndIf
        Inc wIndex                          ' Move up the bins
    Until wIndex >= (cFFT_BinSize / 2)      ' Until half the bins are reached
EndProc

'-------------------------------------------------------------------------
' Find the highest column peak within the FFT array
' Input     : FFT_wSamples
' Output    : Returns the bin that contains the largest value
' Notes     : Only scans the second half of the array because this is where the column frequencies are
'
Proc DTMF_FindColBin(), Byte
    Dim wIndex As Word
    Dim wTemp As Word
    Dim wLargest As Word = 2

    wIndex = (cFFT_BinSize / 2)             ' Start half way up the bins
    Result = 0                              ' Default return if no bin found
    Repeat                                  ' Create loop to scan the bins
        wTemp = FFT_wSamples[wIndex]        ' Get a bin's value
        If wTemp > wLargest Then            ' Is this value greater than the largest found so far?
            wLargest = wTemp                ' Yes. So make this value the largest so far
            Result = wIndex                 ' Make this value the result
        EndIf
        Inc wIndex                          ' Move up the bins
    Until wIndex >= cFFT_BinSize            ' Until half the bins are reached
EndProc

'-------------------------------------------------------------------------
' Decode the FFT bin values into key values
' Input     : None
' Output    : Result holds the key value (in ASCII)
'           : Result is 0 if no valid tones found
' Notes     : None
'
Proc DTMF_GetKey(), Byte
    Dim bColBin As Byte = DTMF_FindColBin()             ' Find the FFT bin containing the column frequency
    Dim bRowBin As Byte = DTMF_FindRowBin()             ' Find the FFT bin containing the row frequency

    Result = 0                                          ' Default to an unknown key value
    Select bRowBin
        Case 44 To 46                                   ' Is the row frequency 697Hz?
            Select bColBin
                Case 76 To 78                           ' Is the column frequency 1209Hz?
                    Result = "1"
                Case 84 To 86                           ' Is the column frequency 1336Hz?
                    Result = "2"
                Case 92 To 94                           ' Is the column frequency 1477Hz?
                    Result = "3"
                Case 102 To 104                         ' Is the column frequency 1633Hz?
                    Result = "A"
            EndSelect
        Case 48 To 51                                   ' Is the row frequency 770Hz?
            Select bColBin
                Case 76 To 78                           ' Is the column frequency 1209Hz?
                    Result = "4"
                Case 84 To 86                           ' Is the column frequency 1336Hz?
                    Result = "5"
                Case 92 To 94                           ' Is the column frequency 1477Hz?
                    Result = "6"
                Case 102 To 104                         ' Is the column frequency 1633Hz?
                    Result = "B"
            EndSelect
        Case 52 To 55                                   ' Is the row frequency 853Hz?
            Select bColBin
                Case 76 To 78                           ' Is the column frequency 1209Hz?
                    Result = "7"
                Case 84 To 86                           ' Is the column frequency 1336Hz?
                    Result = "8"
                Case 92 To 94                           ' Is the column frequency 1477Hz?
                    Result = "9"
                Case 102 To 104                         ' Is the column frequency 1633Hz?
                    Result = "C"
            EndSelect
        Case 57 To 61                                   ' Is the row frequency 941Hz?
            Select bColBin
                Case 76 To 78                           ' Is the column frequency 1209Hz?
                    Result = "*"
                Case 84 To 86                           ' Is the column frequency 1336Hz?
                    Result = "0"
                Case 92 To 94                           ' Is the column frequency 1477Hz?
                    Result = "#"
                Case 102 To 104                         ' Is the column frequency 1633Hz?
                    Result = "D"
            EndSelect
    EndSelect
EndProc

'--------------------------------------------------------------------------------
' Receive and decode DTMF signals received on AN0
' Input     : None
' Output    : Returns the ASCII key value. 0 if no key recognised
' Notes     : None
'
Proc DTMF_Decode(), Byte
    While                                                       ' Create a loop to wait for a DTMF buffer fill
        If FFT_tBufferReady = True Then                         ' Is the FFT buffer ready?
            FFT_ComplexIP(FFT_wSamples, TwiddleCoeff_256, 8)    ' Yes. So perform an FFT of the captured data
            FFT_BitReverseComplex(FFT_wSamples, 8)              ' Bit reverse the complex FFT data
            FFT_SquareMagnitudeCplx(FFT_wSamples, FFT_wSamples, cFFT_BinSize) ' Calculate the amplitude of the bins
            Result = DTMF_GetKey()                              ' Place the decoded key value as the return
            FFT_tBufferReady = False                            ' Indicate to the interrupt that we're finished with the current buffer
            Return                                              ' Exit the procedure
        EndIf
    Wend                                                        ' Do it forever (probably requires a timeout in a real application)
EndProc

'--------------------------------------------------------------------------------
' The main program starts here
'
Main:
    PLL_Setup(43, 2, 2, $0300)                      ' Configure the Oscillator to operate the device at 79.23MHz
    PPS_Output(cOut_Pin_RP3, cOut_Fn_U1TX)          ' Make Pin RB3 U1TX

    FIR_wIndex = 0                                  ' Reset the FIR buffer index
    Clear FIR_wSamples                              ' Clear the FIR_wSamples array

    FFT_tBufferReady = False                        ' Reset the buffer ready flag
    FFT_wIndex = 0                                  ' Reset the FFT buffer index
'
' Initialise the FIR Filter parameters with the buffer and coefficient details
'
    FIR_StructInit(FIRStruct, cCOEFF, cFIR_Taps, FIR_wSamples, cFIR_BufferSize)
    ADC_Start()                                     ' Initialise and start the ADC interrupt
    Timer1_Start()                                  ' Initialise and start the Timer1 interrupt
'
' Decode and transmit serially the received DTMF signals
'
    While                                           ' Create an infinite loop
        '
        ' Wait for a valid key
        '
        DTMF_wZeroCounter = 0                       ' Reset the zero counter
        While                                       ' Create an infinite loop
            DTMF_bKeyValue = DTMF_Decode()          ' Decode a DTMF signal
            Inc DTMF_wZeroCounter                   ' Increment the zero counter
            If DTMF_bKeyValue <> 0 Then Break       ' Exit the loop when a valid key press is found
        Wend                                        ' Wait forever (probably requires a timeout in a real program)
        '
        ' Transmit the ASCII key value. Character "*" will cause a <CR> to be transmitted
        '
        If DTMF_wZeroCounter > 1 Then               ' Was the zero counter greater than 1?
            DTMF_bKeyValue = DTMF_Decode()          ' Yes. So decode the DTMF signal
            If DTMF_bKeyValue = "*" Then            ' Is the key a "*" character?
                HRSOut 13                           ' Yes. So transmit a carriage return
            Else                                    ' Otherwise...
                HRSOut DTMF_bKeyValue               ' Transmit the ASCII key value
            EndIf
        EndIf
    Wend                                            ' Do it forever

'---------------------------------------------------------------------------------------
' Configure for internal 7.37MHz oscillator with PLL
' OSC pins are general purpose I/O
'
    Config FBS = BWRP_WRPROTECT_OFF
    Config FGS = GWRP_OFF
    Config FOSCSEL = FNOSC_FRCPLL, IESO_OFF
    Config FOSC = POSCMD_NONE, OSCIOFNC_ON, IOL1WAY_OFF, FCKSM_CSECME
    Config FWDT = WDTPOST_PS256, WINDIS_OFF, FWDTEN_OFF
    Config FPOR = FPWRT_PWR1, ALTI2C_OFF
    Config FICD = ICS_PGD1, JTAGEN_OFF

